/***************************************************************************/
/* Test program for libSIMD, the Mathematics Library using SIMD code.      */
/* Copyright (C) 2000-2002 by Iain Nicholson  <iain@delphinus.demon.co.uk> */
/* Last modified : 14th February 2002                                      */
/*                                                                         */
/* This file is part of libSIMD.                                           */
/*                                                                         */
/*  libSIMD is free software; you can redistribute it and/or modify it     */
/*  under the terms of the GNU Lesser General Public License as published  */
/*  by the Free Software Foundation; either version 2 of the License, or   */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  libSIMD is distributed in the hope that it will be useful, but WITHOUT */
/*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or  */
/*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public   */
/*  License for more details.                                              */
/*                                                                         */
/*  You should have received a copy of the GNU Lesser General Public       */
/*  License along with libSIMD; if not, write to the Free Software         */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307   */
/*  USA                                                                    */
/***************************************************************************/

#include <math.h>
#include <stdio.h>
#include "simdd.h"

#define PI 3.14159269

/* this will do for now, and is merely to show that things work */

void test_scalard(void);
void test_trigd(void);
void test_vectord(void);
void test_vectord2(void);
void test_vectord3(void);
void test_vectord4(void);
void test_matrixd(void);
void test_matrixd2(void);
void test_matrixd3(void);
void test_matrixd4(void);
void test_complexd(void);
void test_dspd(void);
void test_quatd(void);

void matd_show_2x2(double [2][2]);
void matd_show_3x3(double [3][3]);
void matd_show_4x4(double [4][4]);

int main(void)
{

	printf("libSIMD double-precision tests.\n");

	test_scalard();
	test_trigd();
	test_vectord();
	test_matrixd();
	test_complexd();
	test_dspd();
	test_quatd();

	printf("\n");

	
	return 0;
}

/*
**  End of main()
*/

void test_scalard()
{

	printf("Scalar functions.\n");
	printf("None implemented.\n");
	printf("Done.\n");
	
}

void test_trigd()
{

	double angle, result;
	double k;
	int i;
	
	printf("Trigonometry functions.\n");

	k=PI/180;
	for(i=0; i<360; i+=30)
	{
		angle=(double)i*k;
		femms();
		simd_sind(&angle, &result);
		femms();
		printf("simd_sind(%d)=%f\n",i,(double)result);
	}

	for(i=0; i<360; i+=30)
	{
		angle=(double)i*k;
		femms();
		simd_cosd(&angle, &result);
		femms();
		printf("simd_cosd(%d)=%f\n",i,(double)result);
	}

	for(i=0; i<10; i++)
	{
		angle=(double)i;
		femms();
		simd_sinhd(&angle, &result);
		femms();
		printf("simd_sinhd(%d)=%f\n",i,(double)result);
	}

	for(i=0; i<10; i++)
	{
		angle=(double)i;
		femms();
		simd_coshd(&angle, &result);
		femms();
		printf("simd_coshd(%d)=%f\n",i,(double)result);
	}

	for(i=0; i<10; i++)
	{
		angle=(double)i;
		femms();
		simd_expd(&angle, &result);
		femms();
		printf("simd_expsd(%d)=%f\n",i,(double)result);
	}

	printf("Done.\n");

}

void test_vectord()
{

	printf("Vector functions.\n");

	test_vectord2();
	test_vectord3();
	test_vectord4();
		
	printf("Done.\n");
	
}


void test_vectord2(void)
{

	double a[2], b[2], c[2];
	double x, y;

	
	printf("2-vector functions.\n");
	printf("\n");
	
	printf("assign:\n");
	x=1.0;
	y=2.0;
	printf("x=%f, y=%f\n", (double) x, (double) y);
	femms();
	simd_vecd2_assign(x, y, a);
	femms();
	printf("a[0]=%f, a[1]=%f\n\n", (double) a[0], (double) a[1]);

	printf("copy:\n");
	printf("a[0]=%f, a[1]=%f\n", (double) a[0], (double) a[1]);
	femms();
	simd_vecd2_copy(a, b);
	femms();
	printf("b[0]=%f, b[1]=%f\n\n", (double) b[0], (double) b[1]);
	
	printf("add\n");
	printf("a[0]=%f, a[1]=%f\n", (double) a[0], (double) a[1]);
	b[0]=3.0;
	b[1]=4.0;
	printf("b[0]=%f, b[1]=%f\n", (double) b[0], (double) b[1]);
	femms();
	simd_vecd2_add(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f\n\n", (double) c[0], (double) c[1]);

	printf("sub\n");
	printf("a[0]=%f, a[1]=%f\n", (double) a[0], (double) a[1]);
	printf("b[0]=%f, b[1]=%f\n", (double) b[0], (double) b[1]);
	femms();
	simd_vecd2_sub(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f\n\n", (double) c[0], (double) c[1]);


	printf("scalar product\n");
	printf("a[0]=%f, a[1]=%f\n", (double) a[0], (double) a[1]);
	printf("b[0]=%f, b[1]=%f\n", (double) b[0], (double) b[1]);
	femms();
	simd_vecd2_scalar_product(a, b, &x);
	femms();
	printf("x=%f\n\n", (double) x);

	printf("magnitude\n");
	printf("a[0]=%f, a[1]=%f\n", (double) a[0], (double) a[1]);
	femms();
	simd_vecd2_mag(a, &x);
	femms();
	printf("x=%f\n\n", (double) x);

	printf("multiply elements\n");
	printf("a[0]=%f, a[1]=%f\n", (double) a[0], (double) a[1]);
	printf("b[0]=%f, b[1]=%f\n", (double) b[0], (double) b[1]);
	femms();
	simd_vecd2_mul(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f\n\n", (double) c[0], (double) c[1]);

	printf("divide elements\n");
	printf("a[0]=%f, a[1]=%f\n", (double) a[0], (double) a[1]);
	printf("b[0]=%f, b[1]=%f\n", (double) b[0], (double) b[1]);
	femms();
	simd_vecd2_div(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f\n\n", (double) c[0], (double) c[1]);

	printf("reciprocal of elements\n");
	printf("b[0]=%f, b[1]=%f\n", (double) b[0], (double) b[1]);
	femms();
	simd_vecd2_recip(b, c);
	femms();
	printf("c[0]=%f, c[1]=%f\n\n", (double) c[0], (double) c[1]);	
	
	printf("Done.\n");
	
}

void test_vectord3(void)
{
	double a[3], b[3], c[3];
	double x, y, z;
	
	printf("3-vector functions.\n");
	printf("\n");
	
	printf("assign:\n");
	x=1.0;
	y=2.0;
	z=3.0;
	printf("x=%f, y=%f, z=%f\n", (double) x, (double) y, (double) z);
	femms();
	simd_vecd3_assign(x, y, z, a);
	femms();
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n\n", (double) a[0], (double) a[1], (double) a[2]);

	printf("copy:\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n", (double) a[0], (double) a[1], (double) a[2]);
	femms();
	simd_vecd3_copy(a, b);
	femms();
	printf("b[0]=%f, b[1]=%f, b[2]=%f\n\n", (double) b[0], (double) b[1], (double) b[2]);
	
	printf("add\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n\n", (double) a[0], (double) a[1], (double) a[2]);
	b[0]=4.0;
	b[1]=5.0;
	b[2]=6.0;
	printf("b[0]=%f, b[1]=%f, b[2]=%f\n\n", (double) b[0], (double) b[1], (double) b[2]);
	femms();
	simd_vecd3_add(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f\n\n", (double) c[0], (double) c[1], (double) c[2]);

	printf("sub\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n\n", (double) a[0], (double) a[1], (double) a[2]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f\n\n", (double) b[0], (double) b[1], (double) b[2]);
	femms();
	simd_vecd3_sub(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f\n\n", (double) c[0], (double) c[1], (double) c[2]);


	printf("scalar product\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n\n", (double) a[0], (double) a[1], (double) a[2]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f\n\n", (double) b[0], (double) b[1], (double) b[2]);
	femms();
	simd_vecd3_scalar_product(a, b, &x);
	femms();
	printf("x=%f\n\n", (double) x);

	printf("magnitude\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n\n", (double) a[0], (double) a[1], (double) a[2]);
	femms();
	simd_vecd3_mag(a, &x);
	femms();
	printf("x=%f\n\n", (double) x);

	printf("multiply elements\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n\n", (double) a[0], (double) a[1], (double) a[2]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f\n\n", (double) b[0], (double) b[1], (double) b[2]);
	femms();
	simd_vecd3_mul(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f\n\n", (double) c[0], (double) c[1], (double) c[2]);

	printf("divide elements\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f\n\n", (double) a[0], (double) a[1], (double) a[2]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f\n\n", (double) b[0], (double) b[1], (double) b[2]);
	femms();
	simd_vecd3_div(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f\n\n", (double) c[0], (double) c[1], (double) c[2]);

	printf("reciprocal of elements\n");
	printf("b[0]=%f, b[1]=%f, b[2]=%f\n\n", (double) b[0], (double) b[1], (double) b[2]);
	femms();
	simd_vecd3_recip(b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f\n\n", (double) c[0], (double) c[1], (double) c[2]);

	printf("Done.\n");
	
}

void test_vectord4(void)
{
	double a[4], b[4], c[4];
	double x1, x2, x3, x4;
	
	printf("4-vector functions.\n");
	printf("\n");
	
	printf("assign:\n");
	x1=1.0;
	x2=2.0;
	x3=3.0;
	x4=4.0;
	printf("x1=%f, x2=%f, x3=%f, x4=%f\n", (double) x1, (double) x2, (double) x3, (double) x4);
	femms();
	simd_vecd4_assign(x1, x2, x3, x4, a);
	femms();
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);

	printf("copy:\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);
	femms();
	simd_vecd4_copy(a, b);
	femms();
	printf("b[0]=%f, b[1]=%f, b[2]=%f, b[3]=%f\n\n", (double) b[0], (double) b[1], (double) b[2], (double) b[3]);
	
	printf("add\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);
	b[0]=4.0;
	b[1]=5.0;
	b[2]=6.0;
	printf("b[0]=%f, b[1]=%f, b[2]=%f, b[3]=%f\n\n", (double) b[0], (double) b[1], (double) b[2], (double) b[3]);
	femms();
	simd_vecd4_add(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f, c[3]=%f\n\n", (double) c[0], (double) c[1], (double) c[2], (double) c[3]);

	printf("sub\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f, b[3]=%f\n\n", (double) b[0], (double) b[1], (double) b[2], (double) b[3]);
	femms();
	simd_vecd4_sub(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f, c[3]=%f\n\n", (double) c[0], (double) c[1], (double) c[2], (double) c[3]);


	printf("scalar product\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f, b[3]=%f\n\n", (double) b[0], (double) b[1], (double) b[2], (double) b[3]);
	femms();
	simd_vecd4_scalar_product(a, b, &x1);
	femms();
	printf("x1=%f\n\n", (double) x1);

	printf("magnitude\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);
	femms();
	simd_vecd4_mag(a, &x1);
	femms();
	printf("x1=%f\n\n", (double) x1);

	printf("multiply elements\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f, b[3]=%f\n\n", (double) b[0], (double) b[1], (double) b[2], (double) b[3]);
	femms();
	simd_vecd4_mul(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f, c[3]=%f\n\n", (double) c[0], (double) c[1], (double) c[2], (double) c[3]);

	printf("divide elements\n");
	printf("a[0]=%f, a[1]=%f, a[2]=%f, a[3]=%f\n\n", (double) a[0], (double) a[1], (double) a[2], (double) a[3]);
	printf("b[0]=%f, b[1]=%f, b[2]=%f, b[3]=%f\n\n", (double) b[0], (double) b[1], (double) b[2], (double) b[3]);
	femms();
	simd_vecd4_div(a, b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f, c[3]=%f\n\n", (double) c[0], (double) c[1], (double) c[2], (double) c[3]);

	printf("reciprocal of elements\n");
	printf("b[0]=%f, b[1]=%f, b[2]=%f, b[3]=%f\n\n", (double) b[0], (double) b[1], (double) b[2], (double) b[3]);
	femms();
	simd_vecd4_recip(b, c);
	femms();
	printf("c[0]=%f, c[1]=%f, c[2]=%f, c[3]=%f\n\n", (double) c[0], (double) c[1], (double) c[2], (double) c[3]);

	printf("Done.\n");
	
}

void test_matrixd(void)
{

	printf("Matrix functions.\n");

	test_matrixd2();
	test_matrixd3();
	test_matrixd4();
	 
	printf("Done.\n");
	
}


void test_matrixd2(void)
{
	double a[2][2], b[2][2], c[2][2];
	double p[2], q[2];
	double x;
	int i, j;

	printf("2x2 Matrix functions.\n");
	printf("\n");

	printf("multiply 2x2 x 2x2:\n");
	for(j=0; j<2; j++)
	{
		for(i=0; i<2; i++)
		{
			a[j][i]=(double)((j+1)*(i+1));
		}
	}
	for(j=0; j<2; j++)
	{
		for(i=0; i<2; i++)
		{
			b[j][i]=(double)((j+1)*(i+1));
		}
	}
	printf("a:\n");
	matd_show_2x2(a);
	printf("b:\n");
	matd_show_2x2(b);
	femms();
	simd_matd_2x2_mul(a, b, c);
	femms();
	printf("c:\n");
	matd_show_2x2(c);
	printf("\n");

	printf("determinant:\n");
	printf("a:\n");
	matd_show_2x2(a);
	simd_matd_2x2_det(a, &x);
	printf("x=%f\n\n", (double) x);

	printf("multiply 2x2 x 2x1:\n");
	p[0]=2.0;
	p[1]=4.0;
	printf("a:\n");
	matd_show_2x2(a);
	printf("p:\n");	
	printf("p[0]=%f, p[1]=%f\n", (double) p[0], (double) p[1]);
	femms();
	simd_matd_2x1_mul(a, p, q);
	femms();
	printf("q:\n");	
	printf("q[0]=%f, q[1]=%f\n", (double) q[0], (double) q[1]);
	
	printf("Done.\n");
	
}

void test_matrixd3(void)
{
	double a[3][3], b[3][3], c[3][3];
	double p[3], q[3];
	double x;
	int i, j;
	
	printf("3x3 Matrix functions.\n");
	printf("\n");

	printf("multiply 3x3 x 3x3:\n");
	for(j=0; j<3; j++)
	{
		for(i=0; i<3; i++)
		{
			a[j][i]=(double)((j+1)*(i+1));
		}
	}
	for(j=0; j<3; j++)
	{
		for(i=0; i<3; i++)
		{
			b[j][i]=(double)((j+1)*(i+1));
		}
	}
	printf("a:\n");
	matd_show_3x3(a);
	printf("b:\n");
	matd_show_3x3(b);
	femms();
	simd_matd_3x3_mul(a, b, c);
	femms();
	printf("c:\n");
	matd_show_3x3(c);
	printf("\n");

	printf("determinant:\n");
	printf("a:\n");
	matd_show_3x3(a);
	simd_matd_3x3_det(a, &x);
	printf("x=%f\n\n", (double) x);

	printf("multiply 3x3 x 3x1:\n");
	p[0]=2.0;
	p[1]=4.0;
	p[2]=6.0;
	printf("a:\n");
	matd_show_3x3(a);
	printf("p:\n");	
	printf("p[0]=%f, p[1]=%f, p[2]=%f\n", (double) p[0], (double) p[1], (double) p[2]);
	femms();
	simd_matd_3x1_mul(a, p, q);
	femms();
	printf("q:\n");	
	printf("q[0]=%f, q[1]=%f, q[2]=%f\n", (double) q[0], (double) q[1], (double) q[2]);
			
	printf("Done.\n");
	
}

void test_matrixd4(void)
{
	double a[4][4], b[4][4], c[4][4];
	double p[4], q[4];
	double x;
	int i, j;
	printf("4x4 Matrix functions.\n");
	printf("\n");

	printf("multiply 4x4 x 4x4:\n");
	for(j=0; j<4; j++)
	{
		for(i=0; i<4; i++)
		{
			a[j][i]=(double)((j+1)*(i+1));
		}
	}
	for(j=0; j<4; j++)
	{
		for(i=0; i<4; i++)
		{
			b[j][i]=(double)((j+1)*(i+1));
		}
	}
	printf("a:\n");
	matd_show_4x4(a);
	printf("b:\n");
	matd_show_4x4(b);
	femms();
	simd_matd_4x4_mul(a, b, c);
	femms();
	printf("c:\n");
	matd_show_4x4(c);
	printf("\n");

	printf("multiply 4x4 x 4x1:\n");
	p[0]=2.0;
	p[1]=4.0;
	p[2]=6.0;
	p[3]=8.0;
	printf("a:\n");
	matd_show_4x4(a);
	printf("p:\n");	
	printf("p[0]=%f, p[1]=%f, p[2]=%f, p[3]=%f\n", (double) p[0], (double) p[1], (double) p[2], (double) p[3]);
	femms();
	simd_matd_4x1_mul(a, p, q);
	femms();
	printf("q:\n");	
	printf("q[0]=%f, q[1]=%f, q[2]=%f, q[3]=%f\n", (double) q[0], (double) q[1], (double) q[2], (double) q[3]);

	printf("Done.\n");
	
}

void test_complexd(void)
{

	printf("Complex functions.\n");
	printf("Done.\n");
	
}

void test_dspd(void)
{

	printf("DSP functions.\n");
	printf("None implemented.\n");
	printf("Done.\n");
	
}

void test_quatd(void)
{

	printf("Quaternion functions.\n");
	printf("None implemented.\n");
	printf("Done.\n");
	
}

void matd_show_2x2(double a[2][2])
{
	int i, j;
	for(j=0; j<2; j++)
	{
		for(i=0; i<2; i++)
		{
			printf("%f ",a[j][i]);
		}
		printf("\n");
	}
}

void matd_show_3x3(double a[3][3])
{
	int i, j;
	for(j=0; j<3; j++)
	{
		for(i=0; i<3; i++)
		{
			printf("%f ",a[j][i]);
		}
		printf("\n");
	}
}

void matd_show_4x4(double a[4][4])
{
	int i, j;
	for(j=0; j<4; j++)
	{
		for(i=0; i<4; i++)
		{
			printf("%f ",a[j][i]);
		}
		printf("\n");
	}
}