/***************************************************************************/
/* libSIMD Mathematics Library using SIMD code.                            */
/* Copyright (C) 2000,2001 by Iain Nicholson <iain@delphinus.demon.co.uk>  */
/* Last modified : 7th February 2002                                       */
/*                                                                         */
/* This file is part of libSIMD.                                           */
/*                                                                         */
/*  libSIMD is free software; you can redistribute it and/or modify it     */
/*  under the terms of the GNU Lesser General Public License as published  */
/*  by the Free Software Foundation; either version 2 of the License, or   */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  libSIMD is distributed in the hope that it will be useful, but WITHOUT */
/*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or  */
/*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public   */
/*  License for more details.                                              */
/*                                                                         */
/*  You should have received a copy of the GNU Lesser General Public       */
/*  License along with libSIMD; if not, write to the Free Software         */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307   */
/*  USA                                                                    */
/***************************************************************************/

#ifndef LIBSIMD_MATRIXS
 #include "matrixs.h"
#endif

void simd_mats_2x2_mul(float a[2][2], float b[2][2], float c[2][2])
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the multiplication */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%ebx), %%mm0
     movq 8(%%ebx), %%mm1
	 movq  (%%eax), %%mm6
	 movq 8(%%eax), %%mm7

	 movq  %%mm6, %%mm2
	 movq  %%mm7, %%mm3
	 movq  %%mm6, %%mm4
	 movq  %%mm7, %%mm5

     punpckldq  %%mm6, %%mm2
     punpckldq  %%mm7, %%mm3
     punpckhdq  %%mm6, %%mm4
     punpckhdq  %%mm7, %%mm5

     pfmul  %%mm0, %%mm2
     pfmul  %%mm0, %%mm3
     pfmul  %%mm1, %%mm4
     pfmul  %%mm1, %%mm5

     pfadd  %%mm4, %%mm2
     pfadd  %%mm5, %%mm3

     movq  %%mm2,  (%%ecx)
     movq  %%mm3, 8(%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	/* (loops unolled by hand) ;-> */

	c[0][0]=a[0][0]*b[0][0]+a[0][1]*b[1][0]; c[0][1]=a[0][0]*b[0][1]+a[0][1]*b[1][1];
	c[1][0]=a[1][0]*b[0][0]+a[1][1]*b[1][0]; c[1][1]=a[1][0]*b[0][1]+a[1][1]*b[1][1];


#endif

}

/*
 *  Added 14th November 2001
 */
void simd_mats_2x2_det(float a[2][2], float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b */
	/* load up mmx registers */
	/* perform the determinant calculation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%eax), %%mm0
     movq 8(%%eax), %%mm1
     movq %%mm1, %%mm2
     punpckldq %%mm1, %%mm2
     punpckhdq %%mm2, %%mm1
     pfmul %%mm0, %%mm1
     movq %%mm1, %%mm2
     punpckhdq %%mm2, %%mm2
     pfsub %%mm2, %%mm1
     movd %%mm1, (%%ebx)
     "
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */

	*b=(a[0][0]*a[1][1])-(a[1][0]*a[0][1]);

#endif

}

/*
 *  Added 6th December 2001
 */
void simd_mats_2x1_mul(float a[2][2], float b[2], float c[2])
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the calculation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq (%%eax), %%mm0
     movq 8(%%eax), %%mm1
     movq (%%ebx), %%mm2
     pfmul %%mm2, %%mm0
     pfmul %%mm2, %%mm1
     pfacc %%mm1, %%mm0
     movq %%mm0, (%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */

	c[0]=(a[0][0]*b[0])+(a[0][1]*b[1]);
	c[1]=(a[1][0]*b[0])+(a[1][1]*b[1]);

#endif

}

void simd_mats_3x3_mul(float a[3][3], float b[3][3], float c[3][3])
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the multiplication */
	/* store the results */
	
	__asm__ __volatile__(

    "
	##
	## phase 1
	##
   	movq  (%%eax), %%mm0
	movq   %%mm0,  %%mm1
	movq   %%mm0,  %%mm2

	punpckldq %%mm1,  %%mm0
	punpckhdq %%mm1,  %%mm2

	movq   (%%ebx),%%mm3
	movq 12(%%ebx),%%mm4

	pfmul %%mm3, %%mm0
	pfmul %%mm4, %%mm2

	movq  8(%%eax),%%mm1
	movq  %%mm1, %%mm3
	movq  %%mm1, %%mm4

	punpckldq %%mm1, %%mm3
	punpckhdq %%mm1, %%mm4

	movq 24(%%ebx), %%mm5
	pfmul %%mm5, %%mm3
	pfadd %%mm2, %%mm0
	pfadd %%mm3, %%mm0
	movq %%mm0, (%%ecx)
	
    ##
    ## phase 2
    ##
    movq (%%eax), %%mm0
    movq 8(%%eax), %%mm1
    movq 16(%%eax), %%mm2

    movq %%mm0, %%mm3
    movq %%mm1, %%mm4
    movq %%mm2, %%mm5

    punpckhdq %%mm0, %%mm4
    punpckhdq %%mm1, %%mm5
    punpckhdq %%mm2, %%mm3

    punpckldq %%mm4, %%mm0
    punpckldq %%mm5, %%mm1
    punpckldq %%mm2, %%mm3

    movq 8(%%ebx), %%mm4
    movq %%mm4, %%mm6
    movq (%%ebx), %%mm5
	punpckldq %%mm5, %%mm4
	movq 16(%%ebx), %%mm5
	punpckhdq %%mm6, %%mm5
	movd 24(%%ebx), %%mm7
	movd 32(%%ebx), %%mm6
	punpckldq %%mm7, %%mm6

	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm6
	pfmul %%mm3, %%mm5
	pfadd %%mm4, %%mm6
	pfadd %%mm6, %%mm5	
    
    movq %%mm5, 8(%%ecx)
    
    ##
    ## phase 3
    ##

    movq 12(%%eax), %%mm0
    movq 20(%%eax), %%mm1
    movq %%mm0, %%mm2

    punpckldq %%mm0, %%mm0
    punpckldq %%mm1, %%mm1
    punpckhdq %%mm2, %%mm2

    movq 4(%%ebx), %%mm3
    movq 16(%%ebx), %%mm4
    movq 28(%%ebx), %%mm5

    pfmul %%mm0, %%mm3
    pfmul %%mm1, %%mm5
    pfmul %%mm2, %%mm4

    pfadd %%mm3, %%mm5
    pfadd %%mm5, %%mm4
	movq %%mm4, 16(%%ecx)
	 
    ##
    ## phase 4
    ##

	movq 24(%%eax), %%mm0
	movd 32(%%eax), %%mm1
	movq %%mm0, %%mm2

	punpckldq %%mm0, %%mm0
	punpckldq %%mm1, %%mm1
	punpckhdq %%mm2, %%mm2

	movq (%%ebx), %%mm3
	movq 12(%%ebx), %%mm4
	movq 24(%%ebx), %%mm5

	pfmul %%mm0, %%mm3
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm4

	pfadd %%mm3, %%mm5
	pfadd %%mm5, %%mm4
	movq %%mm4, 24(%%ecx)
    
    ##
    ## phase 5
    ##

    movq 24(%%eax), %%mm0
    movd 32(%%eax), %%mm1
    movq 4(%%ebx), %%mm2
    movq 16(%%ebx), %%mm3
    movq 28(%%ebx), %%mm4

    punpckhdq %%mm3, %%mm2
    punpckhdq %%mm4, %%mm4

    pfmul %%mm0, %%mm2
    pfmul %%mm1, %%mm4
    pfacc %%mm5, %%mm2
    pfadd %%mm2, %%mm4
    movd %%mm4, 32(%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	/* I may decide to implement this as a couple of loops */

	c[0][0]=a[0][0]*b[0][0]+a[0][1]*b[1][0]+a[0][2]*b[2][0];
	c[0][1]=a[0][0]*b[0][1]+a[0][1]*b[1][1]+a[0][2]*b[2][1];
	c[0][2]=a[0][0]*b[0][2]+a[0][1]*b[1][2]+a[0][2]*b[2][2];
	
	c[1][0]=a[1][0]*b[0][0]+a[1][1]*b[1][0]+a[1][2]*b[2][0];
	c[1][1]=a[1][0]*b[0][1]+a[1][1]*b[1][1]+a[1][2]*b[2][1];
	c[1][2]=a[1][0]*b[0][2]+a[1][1]*b[1][2]+a[1][2]*b[2][2];
	
	c[2][0]=a[2][0]*b[0][0]+a[2][1]*b[1][0]+a[2][2]*b[2][0];
	c[2][1]=a[2][0]*b[0][1]+a[2][1]*b[1][1]+a[2][2]*b[2][1];
	c[2][2]=a[2][0]*b[0][2]+a[2][1]*b[1][2]+a[2][2]*b[2][2];
	
#endif

}


/*
 *  Added 7th February 2002
 *  I'm not particularly proud of this.
 *  Don't try this at home kids!
 */

void simd_mats_3x3_det(float a[3][3], float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b */
	/* load up mmx registers */
	/* perform the multiplication */
	/* store the results */
	
	__asm__ __volatile__(

    "
	##
	## phase 1
	##
	movq   16(%%eax), %%mm0
	movq   28(%%eax), %%mm1
	movq      %%mm1,  %%mm2
	punpckldq %%mm1,  %%mm2
	punpckhdq %%mm2,  %%mm1
	pfmul     %%mm0,  %%mm1
	movq      %%mm1,  %%mm2
	punpckhdq %%mm2,  %%mm2
	pfsub     %%mm2,  %%mm1
	##
	## phase 2
	##
	movd    8(%%eax), %%mm2
	movq   20(%%eax), %%mm3
	movd   32(%%eax), %%mm4
	punpckldq %%mm3,  %%mm4
	punpckhdq %%mm3,  %%mm2
	pfmul     %%mm4,  %%mm2
	movq      %%mm2,  %%mm4
	punpckhdq %%mm4,  %%mm4
	pfsub     %%mm4,  %%mm2
	##
	## phase 3
	##
	movq   12(%%eax), %%mm3
	movq   24(%%eax), %%mm4
	movq      %%mm4,  %%mm5
	punpckldq %%mm4,  %%mm5
	punpckhdq %%mm5,  %%mm4
	pfmul     %%mm3,  %%mm5
	##
	## phase 4
	##
	movd     (%%eax), %%mm5
	movd    4(%%eax), %%mm6
	movd    8(%%eax), %%mm7
	pfmul     %%mm1,  %%mm5
	pfmul     %%mm2,  %%mm6
	pfmul     %%mm4,  %%mm7
	pfsub     %%mm6,  %%mm5
	pfadd     %%mm7,  %%mm5
	movd      %%mm5, (%%ebx) # store result
     "
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */

	*b = (a[0][0]* (a[1][1]*a[2][2] - a[2][1]*a[1][2]))
	    -(a[0][1]* (a[1][0]*a[2][2] - a[2][0]*a[1][2]))
	    +(a[0][2]* (a[1][0]*a[2][1] - a[2][0]*a[1][1]));
	
#endif

}


/*
 *  Added 6th December 2001
 */
void simd_mats_3x1_mul(float a[3][3], float b[3], float c[3])
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b %ecx=c */
	/* load up mmx registers */
	/* perform the calculation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq (%%ebx), %%mm0
     movd 8(%%ebx), %%mm1
	 ##
	 ## row 1
	 ##
     movq (%%eax), %%mm2
     movd 8(%%eax), %%mm3
     pfmul %%mm0, %%mm2
     pfmul %%mm1, %%mm3
     pfadd %%mm3, %%mm2
     ##
     ## row 2
     ##
     movq 12(%%eax), %%mm3
     movd 20(%%eax), %%mm4
     pfmul %%mm0, %%mm3
     pfmul %%mm1, %%mm4
     pfadd %%mm4, %%mm3
     ##
     ## convolve two results and store
     ##
     pfacc %%mm3, %%mm2
     movq %%mm2, (%%ecx)
     ##
     ## row 3
     ##
     movq 24(%%eax), %%mm2
     movd 32(%%eax), %%mm3
     pfmul %%mm0, %%mm2
     pfmul %%mm1, %%mm3
     pfadd %%mm3, %%mm2
     pfacc %%mm2, %%mm2
     movd %%mm2, 8(%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */

	c[0]=(a[0][0]*b[0])+(a[0][1]*b[1])+(a[0][2]*b[2]);
	c[1]=(a[1][0]*b[0])+(a[1][1]*b[1])+(a[1][2]*b[2]);
	c[2]=(a[2][0]*b[0])+(a[2][1]*b[1])+(a[2][2]*b[2]);

#endif

}

void simd_mats_4x4_mul(float a[4][4], float b[4][4], float c[4][4])
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	
	__asm__ __volatile__(

    "
	##
	## phase 1A
	##
	## Set up the top row of a
	##
	movq  (%%eax), %%mm0
	movq 8(%%eax), %%mm2
	movq %%mm0, %%mm1
	movq %%mm2, %%mm3
	punpckldq %%mm0, %%mm0
	punpckldq %%mm2, %%mm2
	punpckhdq %%mm1, %%mm1
	punpckhdq %%mm3, %%mm3

	##
	## get the LHS of b
	##
	movq   (%%ebx), %%mm4
	movq 16(%%ebx), %%mm5
	movq 32(%%ebx), %%mm6
	movq 48(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (1A)
	##
	movq %%mm7, (%%ecx)

	##
	## phase 1B
	##
	##
	## get the RHS of b
	##
	movq  8(%%ebx), %%mm4
	movq 24(%%ebx), %%mm5
	movq 40(%%ebx), %%mm6
	movq 56(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (1B)
	##
	movq %%mm7, 8(%%ecx)

	##
	## phase 2A
	##
	## Set up the second row of a
	##
	movq 16(%%eax), %%mm0
	movq 24(%%eax), %%mm2
	movq %%mm0, %%mm1
	movq %%mm2, %%mm3
	punpckldq %%mm0, %%mm0
	punpckldq %%mm2, %%mm2
	punpckhdq %%mm1, %%mm1
	punpckhdq %%mm3, %%mm3

	##
	## get the LHS of b
	##
	movq   (%%ebx), %%mm4
	movq 16(%%ebx), %%mm5
	movq 32(%%ebx), %%mm6
	movq 48(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (2A)
	##
	movq %%mm7, 16(%%ecx)

	##
	## phase 2B
	##
	##
	## get the RHS of b
	##
	movq  8(%%ebx), %%mm4
	movq 24(%%ebx), %%mm5
	movq 40(%%ebx), %%mm6
	movq 56(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (2B)
	##
	movq %%mm7, 24(%%ecx)

	##
	## phase 3A
	##
	## Set up the third row of a
	##
	movq 32(%%eax), %%mm0
	movq 40(%%eax), %%mm2
	movq %%mm0, %%mm1
	movq %%mm2, %%mm3
	punpckldq %%mm0, %%mm0
	punpckldq %%mm2, %%mm2
	punpckhdq %%mm1, %%mm1
	punpckhdq %%mm3, %%mm3

	##
	## get the LHS of b
	##
	movq   (%%ebx), %%mm4
	movq 16(%%ebx), %%mm5
	movq 32(%%ebx), %%mm6
	movq 48(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (3A)
	##
	movq %%mm7, 32(%%ecx)

	##
	## phase 3B
	##
	##
	## get the RHS of b
	##
	movq  8(%%ebx), %%mm4
	movq 24(%%ebx), %%mm5
	movq 40(%%ebx), %%mm6
	movq 56(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (3B)
	##
	movq %%mm7, 40(%%ecx)

	##
	## phase 4A
	##
	## Set up the fourth row of a
	##
	movq 48(%%eax), %%mm0
	movq 56(%%eax), %%mm2
	movq %%mm0, %%mm1
	movq %%mm2, %%mm3
	punpckldq %%mm0, %%mm0
	punpckldq %%mm2, %%mm2
	punpckhdq %%mm1, %%mm1
	punpckhdq %%mm3, %%mm3

	##
	## get the LHS of b
	##
	movq   (%%ebx), %%mm4
	movq 16(%%ebx), %%mm5
	movq 32(%%ebx), %%mm6
	movq 48(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (4A)
	##
	movq %%mm7, 48(%%ecx)

	##
	## phase 4B
	##
	##
	## get the RHS of b
	##
	movq  8(%%ebx), %%mm4
	movq 24(%%ebx), %%mm5
	movq 40(%%ebx), %%mm6
	movq 56(%%ebx), %%mm7

	##
	## calcultae
	##
	pfmul %%mm0, %%mm4
	pfmul %%mm1, %%mm5
	pfmul %%mm2, %%mm6
	pfmul %%mm3, %%mm7

	pfadd %%mm4, %%mm5
	pfadd %%mm5, %%mm6
	pfadd %%mm6, %%mm7

	##
	## store results (4B)
	##
	movq %%mm7,56(%%ecx)
	
    "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	/* I may decide to implement this as a couple of loops */

	c[0][0]=a[0][0]*b[0][0]+a[0][1]*b[1][0]+a[0][2]*b[2][0]+a[0][3]*b[3][0];
	c[0][1]=a[0][0]*b[0][1]+a[0][1]*b[1][1]+a[0][2]*b[2][1]+a[0][3]*b[3][1];
	c[0][2]=a[0][0]*b[0][2]+a[0][1]*b[1][2]+a[0][2]*b[2][2]+a[0][3]*b[3][2];
	c[0][3]=a[0][0]*b[0][3]+a[0][1]*b[1][3]+a[0][2]*b[2][3]+a[0][3]*b[3][3];
	c[1][0]=a[1][0]*b[0][0]+a[1][1]*b[1][0]+a[1][2]*b[2][0]+a[1][3]*b[3][0];
	c[1][1]=a[1][0]*b[0][1]+a[1][1]*b[1][1]+a[1][2]*b[2][1]+a[1][3]*b[3][1];
	c[1][2]=a[1][0]*b[0][2]+a[1][1]*b[1][2]+a[1][2]*b[2][2]+a[1][3]*b[3][2];
	c[1][3]=a[1][0]*b[0][3]+a[1][1]*b[1][3]+a[1][2]*b[2][3]+a[1][3]*b[3][3];
	c[2][0]=a[2][0]*b[0][0]+a[2][1]*b[1][0]+a[2][2]*b[2][0]+a[2][3]*b[3][0];
	c[2][1]=a[2][0]*b[0][1]+a[2][1]*b[1][1]+a[2][2]*b[2][1]+a[2][3]*b[3][1];
	c[2][2]=a[2][0]*b[0][2]+a[2][1]*b[1][2]+a[2][2]*b[2][2]+a[2][3]*b[3][2];
	c[2][3]=a[2][0]*b[0][3]+a[2][1]*b[1][3]+a[2][2]*b[2][3]+a[2][3]*b[3][3];
	c[3][0]=a[3][0]*b[0][0]+a[3][1]*b[1][0]+a[3][2]*b[2][0]+a[3][3]*b[3][0];
	c[3][1]=a[3][0]*b[0][1]+a[3][1]*b[1][1]+a[3][2]*b[2][1]+a[3][3]*b[3][1];
	c[3][2]=a[3][0]*b[0][2]+a[3][1]*b[1][2]+a[3][2]*b[2][2]+a[3][3]*b[3][2];
	c[3][3]=a[3][0]*b[0][3]+a[3][1]*b[1][3]+a[3][2]*b[2][3]+a[3][3]*b[3][3];

#endif

}

void simd_mats_4x1_mul(float a[4][4], float b[4], float c[4])
{
#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the multiplication */
	/* store the results */
	
	__asm__ __volatile__(

    "movq (%%ebx), %%mm0
     movq 8(%%ebx),%%mm1
     movq (%%eax), %%mm2
     movq 8(%%eax),%%mm3
     pfmul %%mm0, %%mm2
     pfmul %%mm1, %%mm3
     pfadd %%mm2,%%mm3
     
     movq 16(%%eax), %%mm4
     movq 24(%%eax),%%mm5
     pfmul %%mm0, %%mm4
     pfmul %%mm1, %%mm5
     pfadd %%mm4,%%mm5

     pfacc %%m5, %%mm3
     movq %%mm3, (%%ecx)

	 movq 32(%%eax), %%mm2
     movq 40(%%eax),%%mm3
     pfmul %%mm0, %%mm2
     pfmul %%mm1, %%mm3
     pfadd %%mm2,%%mm3
     
     movq 48(%%eax), %%mm4
     movq 56(%%eax),%%mm5
     pfmul %%mm0, %%mm4
     pfmul %%mm1, %%mm5
     pfadd %%mm4,%%mm5

     pfacc %%m5, %%mm3
     movq %%mm3, 8(%%ecx)     
    "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	/* drunk and listening to Ozzy : */

	c[0]=(a[0][0]*b[0])+(a[0][1]*b[1])+(a[0][2]*b[2])+(a[0][3]*b[3]);
	c[1]=(a[1][0]*b[0])+(a[1][1]*b[1])+(a[1][2]*b[2])+(a[1][3]*b[3]);
	c[2]=(a[2][0]*b[0])+(a[2][1]*b[1])+(a[2][2]*b[2])+(a[2][3]*b[3]);
	c[3]=(a[3][0]*b[0])+(a[3][1]*b[1])+(a[3][2]*b[2])+(a[3][3]*b[3]);


#endif

}