/***************************************************************************/
/* libSIMD Mathematics Library using SIMD code.                            */
/* Copyright (C) 2000,2001 by Iain Nicholson <iain@delphinus.demon.co.uk>  */
/* Last modified : 14th February 2002                                      */
/*                                                                         */
/* This file is part of libSIMD.                                           */
/*                                                                         */
/*  libSIMD is free software; you can redistribute it and/or modify it     */
/*  under the terms of the GNU Lesser General Public License as published  */
/*  by the Free Software Foundation; either version 2 of the License, or   */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  libSIMD is distributed in the hope that it will be useful, but WITHOUT */
/*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or  */
/*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public   */
/*  License for more details.                                              */
/*                                                                         */
/*  You should have received a copy of the GNU Lesser General Public       */
/*  License along with libSIMD; if not, write to the Free Software         */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307   */
/*  USA                                                                    */
/***************************************************************************/

/*
 * NOTE : THESE FUNCTIONS ARE A FIRST TRY. THEY ARE NOT ACCURATE!
 */

#ifndef LIBSIMD_TRIGS
 #include "trigonometrys.h"
#endif

/*****************************************************************************/
/* simd_sins() : calculate the single-precision sin of an angle (in radians) */
/* using a Taylor series.                                                    */
/*****************************************************************************/

void simd_sins(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b (result) */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	/* NOTE : this is not very accurate !!!!! */

	__asm__ __volatile__(

    "movd (%%eax), %%mm0				# the angle in radians, result
	 movq %%mm0, %%mm7
	 movq %%mm0, %%mm1			    	# numerator
	 movl $0x3f800000, %%edx			# 1.0
	 movd %%edx, %%mm2					# 1.0 (denominator)
	 movq %%mm2, %%mm3					# for the factorial computation
	 movq %%mm2, %%mm4					# a handy constant for incrementing

	 movl $6, %%ecx						# no. of iterations

     simd_sins_l1:
     	pfmul %%mm7, %%mm1					# numerator
    	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfsub %%mm6, %%mm0					# cumulative result


     	pfmul %%mm7, %%mm1					# numerator
     	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result

		dec %%ecx
		cmp $0,%%ecx
    jne simd_sins_l1
    movd %%mm0, (%%ebx)				# store result
    #
    # ..and if you believe that, you'll believe anything! :-)
    #
     "
	 :
	 : "a" (a), "b" (b)
	 : "ecx", "edx", "memory");
#else
	int n;
	float result, num, den, f;
	/* this is what it does in C : */

	result=*a;
	num=result;
	den=1.0;
	f=1.0;

	for(n=0; n<=5; n++)
	{
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result-=num/den;
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result+=num/den;
	}
	*b=result;
#endif
}

void simd_coss(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b (result) */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	/* NOTE : this is not very accurate !!!!! */

	__asm__ __volatile__(

    "movd (%%eax), %%mm7				# the angle in radians
	 movl $0x3f800000, %%edx			# 1.0
	 movd %%edx, %%mm1			    	# numerator
	 movq %%mm1, %%mm0 
	 movq %%mm1, %%mm2					# 1.0 (denominator)
	 movq %%mm2, %%mm3
	 pfsub %%mm2, %%mm3					# 0.0 for the factorial computation
	 movq %%mm2, %%mm4					# a handy constant for incrementing

	 movl $6, %%ecx						# no. of iterations

     simd_coss_l1:
     	pfmul %%mm7, %%mm1					# numerator
    	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfsub %%mm6, %%mm0					# cumulative result


     	pfmul %%mm7, %%mm1					# numerator
     	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result

		dec %%ecx
		cmp $0,%%ecx
    jne simd_coss_l1
    movd %%mm0, (%%ebx)				# store result
    #
    # ..and if you believe that, you'll believe anything! :-)
    #
     "
	 :
	 : "a" (a), "b" (b)
	 : "ecx", "edx", "memory");
#else
	int n;
	float result, num, den, f;
	/* this is what it does in C : */

	result=1;
	num=1.0;
	den=1.0;
	f=0.0;

	for(n=0; n<=5; n++)
	{
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result-=num/den;
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result+=num/den;
	}
	*b=result;
#endif
}

void simd_tans(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b (result) */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	/* NOTE : this is not very accurate !!!!! */

/*	__asm__ __volatile__(

    "movd (%%eax), %%mm7				# the angle in radians
	 movl $0x3f800000, %%edx			# 1.0
	 movd %%edx, %%mm1			    	# numerator
	 movq %%mm1, %%mm0 
	 movq %%mm1, %%mm2					# 1.0 (denominator)
	 movq %%mm2, %%mm3
	 pfsub %%mm2, %%mm3					# 0.0 for the factorial computation
	 movq %%mm2, %%mm4					# a handy constant for incrementing

	 movl $6, %%ecx						# no. of iterations

     simd_coss_l1:
     	pfmul %%mm7, %%mm1					# numerator
    	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfsub %%mm6, %%mm0					# cumulative result


     	pfmul %%mm7, %%mm1					# numerator
     	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result

		dec %%ecx
		cmp $0,%%ecx
    jne simd_coss_l1
    movd %%mm0, (%%ebx)				# store result
    #
    # ..and if you believe that, you'll believe anything! :-)
    #
     "
	 :
	 : "a" (a), "b" (b)
	 : "ecx", "edx", "memory");*/
#else
	int n;
	float result, num, den, f;
	/* this is what it does in C : */

/*	result=1;
	num=1.0;
	den=1.0;
	f=0.0;

	for(n=0; n<=5; n++)
	{
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result-=num/den;
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result+=num/den;
	}
	*b=result;*/
#endif
}


/* inverses */


/* hyperbolic functions */

void simd_sinhs(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b (result) */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	/* NOTE : this is not very accurate !!!!! */

	__asm__ __volatile__(

    "movd (%%eax), %%mm0				# the angle in radians, result
	 movq %%mm0, %%mm7
	 movq %%mm0, %%mm1			    	# numerator
	 movl $0x3f800000, %%edx			# 1.0
	 movd %%edx, %%mm2					# 1.0 (denominator)
	 movq %%mm2, %%mm3					# for the factorial computation
	 movq %%mm2, %%mm4					# a handy constant for incrementing

	 movl $6, %%ecx						# no. of iterations

     simd_sinhs_l1:
     	pfmul %%mm7, %%mm1					# numerator
    	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result


     	pfmul %%mm7, %%mm1					# numerator
     	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result

		dec %%ecx
		cmp $0,%%ecx
    jne simd_sinhs_l1
    movd %%mm0, (%%ebx)				# store result
    #
    # ..and if you believe that, you'll believe anything! :-)
    #
     "
	 :
	 : "a" (a), "b" (b)
	 : "ecx", "edx", "memory");
#else
	int n;
	float result, num, den, f;
	/* this is what it does in C : */

	result=*a;
	num=result;
	den=1.0;
	f=1.0;

	for(n=0; n<=5; n++)
	{
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result+=num/den;
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result+=num/den;
	}
	*b=result;
#endif
}

void simd_coshs(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b (result) */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	/* NOTE : this is not very accurate !!!!! */

	__asm__ __volatile__(

    "movd (%%eax), %%mm7				# the angle in radians
	 movl $0x3f800000, %%edx			# 1.0
	 movd %%edx, %%mm1			    	# numerator
	 movq %%mm1, %%mm0 
	 movq %%mm1, %%mm2					# 1.0 (denominator)
	 movq %%mm2, %%mm3
	 pfsub %%mm2, %%mm3					# 0.0 for the factorial computation
	 movq %%mm2, %%mm4					# a handy constant for incrementing

	 movl $6, %%ecx						# no. of iterations

     simd_coshs_l1:
     	pfmul %%mm7, %%mm1					# numerator
    	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result


     	pfmul %%mm7, %%mm1					# numerator
     	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# factorial
     	pfadd %%mm4, %%mm3					# add 1.0
		pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result

		dec %%ecx
		cmp $0,%%ecx
    jne simd_coshs_l1
    movd %%mm0, (%%ebx)				# store result
    #
    # ..and if you believe that, you'll believe anything! :-)
    #
     "
	 :
	 : "a" (a), "b" (b)
	 : "ecx", "edx", "memory");
#else
	int n;
	float result, num, den, f;
	/* this is what it does in C : */

	result=1;
	num=1.0;
	den=1.0;
	f=0.0;

	for(n=0; n<=5; n++)
	{
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result+=num/den;
		num*=*a;
		num*=*a;
		f++;
		den*=f;
		f++;
		den*=f;
		result+=num/den;
	}
	*b=result;
#endif
}

/* hyperbolic inverses */



/* e^x */

void simd_exps(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b (result) */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	/* NOTE : this is not very accurate !!!!! */

	__asm__ __volatile__(

    "movd (%%eax), %%mm7				# the angle in radians
	 movl $0x3f800000, %%edx			# 1.0
	 movd %%edx, %%mm1			    	# numerator
	 movq %%mm1, %%mm0 
	 movq %%mm1, %%mm2					# 1.0 (denominator)
	 movq %%mm2, %%mm3
	 pfsub %%mm2, %%mm3					# 0.0 for the factorial computation
	 movq %%mm2, %%mm4					# a handy constant for incrementing

	 movl $12, %%ecx						# no. of iterations

     simd_exps_l1:
    	pfmul %%mm7, %%mm1					# new numerator
     	pfadd %%mm4, %%mm3					# add 1.0
     	pfmul %%mm3, %%mm2					# new factorial

		movq %%mm2, %%mm6
     	pfrcp %%mm6, %%mm5					#
     	pfrcpit1 %%mm5, %%mm6				#
    	pfrcpit2 %%mm5, %%mm6				# 1/denominator

 	    pfmul %%mm1, %%mm6					# num * 1/den
 	    pfadd %%mm6, %%mm0					# cumulative result

		dec %%ecx
		cmp $0,%%ecx
    jne simd_exps_l1
    movd %%mm0, (%%ebx)				# store result
    #
    # ..and if you believe that, you'll believe anything! :-)
    #
     "
	 :
	 : "a" (a), "b" (b)
	 : "ecx", "edx", "memory");
#else
	int n;
	float result, num, den, f;
	/* this is what it does in C : */

	result=1;
	num=1.0;
	den=1.0;
	f=0.0;

	for(n=0; n<=12; n++)
	{
		num*=*a;
		f++;
		den*=f;
		result+=num/den;
	}
	*b=result;
#endif
}



