/***************************************************************************/
/* libSIMD Mathematics Library using SIMD code.                            */
/* Copyright (C) 2000,2001 by Iain Nicholson <iain@delphinus.demon.co.uk>  */
/* Last modified : 14th February 2002                                      */
/*                                                                         */
/* This file is part of libSIMD.                                           */
/*                                                                         */
/*  libSIMD is free software; you can redistribute it and/or modify it     */
/*  under the terms of the GNU Lesser General Public License as published  */
/*  by the Free Software Foundation; either version 2 of the License, or   */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  libSIMD is distributed in the hope that it will be useful, but WITHOUT */
/*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or  */
/*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public   */
/*  License for more details.                                              */
/*                                                                         */
/*  You should have received a copy of the GNU Lesser General Public       */
/*  License along with libSIMD; if not, write to the Free Software         */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307   */
/*  USA                                                                    */
/***************************************************************************/

#ifndef LIBSIMD_VECTORS
 #include "vectors.h"
#endif

#include <math.h>

/***************************************************************************/
/* 10th December 2001: Now we really tidy this up aiming for Version 0.0.1 */
/***************************************************************************/

/*************/
/* 2-vectors */
/*************/

void simd_vecs2_assign(float a, float b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movd  %%eax, %%mm0
     movd  %%ebx, %%mm1
     pfacc %%mm1, %%mm0
     movq  %%mm0, (%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a;
	c[1]=b;

#endif
}

void simd_vecs2_copy(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq (%%eax), %%mm0
     movq  %%mm0, (%%ebx)
     "
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	b[0]=a[0];
	b[1]=a[1];

#endif

}

void simd_vecs2_add(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%eax), %%mm0
     pfadd (%%ebx), %%mm0
     movq   %%mm0, (%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]+b[0];
	c[1]=a[1]+b[1];

#endif

}

void simd_vecs2_sub(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%eax), %%mm0
     pfsub (%%ebx), %%mm0
     movq   %%mm0, (%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]-b[0];
	c[1]=a[1]-b[1];

#endif

}

void simd_vecs2_scalar_product(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%eax), %%mm0
     pfmul (%%ebx), %%mm0
     pfacc  %%mm0,  %%mm0
     movd   %%mm0, (%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	*c=(a[0]*b[0])+(a[1]*b[1]);

#endif

}


/*void simd_vecs2_vecstor_product(float *, float *, float *, float *, float *, float *);*/

void simd_vecs2_mag(float *a, float *b)
{
#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = result */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq      (%%eax), %%mm0
     pfmul      %%mm0,  %%mm0
     pfacc      %%mm0,  %%mm0
     pfsqrt     %%mm0,  %%mm1
     pfsqrtit1  %%mm1,  %%mm0
     pfsqrtit2  %%mm1,  %%mm0
     movd       %%mm0, (%%ebx)
     "
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	*b=sqrt((double)(a[0]*a[0]+a[1]*a[1]));

#endif

}



void simd_vecs2_mul(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq (%%eax), %%mm0
     pfmul (%%ebx), %%mm0
     movq %%mm0, (%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]*b[0];
	c[1]=a[1]*b[1];

#endif

}

void simd_vecs2_div(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b, %ecx = c */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movd    (%%eax), %%mm0
     movd   4(%%eax), %%mm1
     pfrcp    %%mm0,  %%mm2
     pfrcpit1 %%mm2,  %%mm0
     pfrcpit2 %%mm2,  %%mm0
     pfrcp    %%mm1,  %%mm3
     pfrcpit1 %%mm3,  %%mm1
     pfrcpit2 %%mm3,  %%mm1
     movd    (%%ebx), %%mm2
     movd   4(%%ebx), %%mm3
     pfmul    %%mm2,  %%mm0
     pfmul    %%mm3,  %%mm1
     movd     %%mm0, (%%ecx)
     movd     %%mm1,4(%%ecx)
     "
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]/b[0];
	c[1]=a[1]/b[1];

#endif

}

void simd_vecs2_recip(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movd    (%%eax), %%mm0
     movd   4(%%eax), %%mm1
     pfrcp    %%mm0,  %%mm2
     pfrcpit1 %%mm2,  %%mm0
     pfrcpit2 %%mm2,  %%mm0
     pfrcp    %%mm1,  %%mm3
     pfrcpit1 %%mm3,  %%mm1
     pfrcpit2 %%mm3,  %%mm1
     movd     %%mm0, (%%ebx)
     movd     %%mm1,4(%%ebx)
     "
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	b[0]=1/a[0];
	b[1]=1/a[1];

#endif

}

/*void simd_vecs2_neg();*/

/*************/
/* 3-vectors */
/*************/

/**********************************************************************/
/* Assign floating point numbers a, b and c to vector pointed to by v */
/* using 3DNow! (of any benefit?)                                     */
/**********************************************************************/

void simd_vecs3_assign(float a, float b, float c, float *v)
{


#ifdef USE_3DNOW

	/* set up the pointers : %eax = v, %edx = &a */
	/* move into MMX registers a,b and c*/
	/* assign to v */
	
	__asm__ __volatile__(

	"movd  %%eax,   %%mm0
	 movd  %%ebx,   %%mm1
	 movd  %%ecx,   %%mm2
	 pfacc %%mm1,   %%mm0
	 movq  %%mm0,  (%%edx)
	 movd  %%mm2, 8(%%edx)
	"
	 :
	 : "a" (a), "b" (b), "c" (c), "d", (v)
	 : "memory");

#else

	/* in C : */
	v[0]=a;
	v[1]=b;
	v[2]=c;

#endif

} 

/*******************************************/
/* Copy vector a into vetor b using 3DNow! */
/*******************************************/

void simd_vecs3_copy(float *a, float *b)
{

#ifdef USE_3DNOW

	/* set up the pointers : %eax = a, %edx = b */
	/* move elements of a into MMX registers */
	/* assign to b */
	__asm__ __volatile__(
	"movq  (%%eax),  %%mm0
	 movq   %%mm0,  (%%ebx)
	 movd 8(%%eax),  %%mm1
	 movd   %%mm1, 8(%%ebx)"
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* in C : */
	b[0]=a[0];
	b[1]=a[1];
	b[2]=a[2];

#endif

}

/*********************************************************/
/* Sum two 3-element floating point vectors using 3DNow! */
/* *a , *b : pointers to vectors to be added             */
/* *c      : pointer to vector in which to store result  */
/*********************************************************/

void simd_vecs3_add(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = c, %edx = a, %ecx = b */
	/* load up mmx registers */
	/* perform the addition */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%eax),  %%mm0
     movq  (%%ebx),  %%mm2
     pfadd  %%mm2,   %%mm0
	 movd 8(%%eax),  %%mm1
	 movd 8(%%ebx),  %%mm3
	 pfadd  %%mm3,   %%mm1
	 movq   %%mm0,  (%%ecx)
	 movd   %%mm1, 8(%%ecx)"
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]+b[0];
	c[1]=a[1]+b[1];
	c[2]=a[2]+b[2];

#endif

}


/**************************************************************/
/* Subtract two 3-element floating point vectors using 3DNow! */
/* *a , *b : pointers to vectors to be subtracted             */
/* *c      : pointer to vector in which to store result       */
/**************************************************************/

void simd_vecs3_sub(float *a, float *b, float *c)
{

#ifdef USE_3DNOW

	/* set up pointers : %eax = c, %edx = a, %ecx = b */
	/* load up mmx registers */
	/* perform the addition */
	/* store the results */
	
	__asm__ __volatile__(
	"movq  (%%eax), %%mm0
	 movq  (%%ebx), %%mm2
	 pfsub  %%mm2,  %%mm0
	 movd 8(%%eax), %%mm1
	 movd 8(%%ebx), %%mm3
	 pfsub  %%mm3,  %%mm1
	 movq   %%mm0, (%%ecx)
	 movd   %%mm1,8(%%ecx)"
	 :
	 : "a" (c), "d" (a), "c" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]-b[0];
	c[1]=a[1]-b[1];
	c[2]=a[2]-b[2];

#endif

}


/**********************************************/
/* perform two scalar products on 2 3-vectors */
/**********************************************/

void simd_vecs3_scalar_product(float *a, float *b, float *c)
{

	#ifdef USE_3DNOW

	__asm__ __volatile(
	"movq   (%%eax), %%mm0
	 movq   (%%ebx), %%mm2
	 movd  8(%%eax), %%mm1
	 movd  8(%%ebx), %%mm3
	 pfmul   %%mm2,  %%mm0		# double
	 pfmul   %%mm3,  %%mm1      # single
	 
	 pfacc %%mm0, %%mm0     # accumulate the double
	 pfadd %%mm2, %%mm0     # add to the single

	 movd %%mm0, (%%ecx)    # result in c
	"
	:
	: "a" (a), "b" (b), "c" (c)
	: "memory");

	#else

	*c=(a[0]*b[0])+(a[1]*b[1])+(a[2]*b[2]);

	#endif

}

void simd_vecs3_mag(float *a, float *b)
{
#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = result */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq      (%%eax), %%mm0
     movd     8(%%eax), %%mm1
     pfmul      %%mm0,  %%mm0
     pfmul      %%mm1,  %%mm1
     pfacc      %%mm0,  %%mm0
     pfadd      %%mm1,  %%mm0
     pfsqrt     %%mm0,  %%mm1
     pfsqrtit1  %%mm1,  %%mm0
     pfsqrtit2  %%mm1,  %%mm0
     movd       %%mm0, (%%ebx)
     "
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	*b=sqrt((double)(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]));

#endif

}


/**********************************************************************/
/* Multiply six floating-point numbers using 3DNow!                  */
/* *a , *b : pointers to arrays of 4 floats containing the input data */
/* *c      : pointer to array of 4 floats in which to store result    */
/**********************************************************************/

void simd_vecs3_mul(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = c, %edx = a, %ecx = b */
	/* load up mmx registers */
	/* perform the multiplication */
	/* store the results */
	
	__asm__ __volatile__(

    "movq   (%%eax), %%mm0
     movq   (%%ebx), %%mm2
	 movd  8(%%eax), %%mm1
	 movd  8(%%ebx), %%mm3
	 pfmul   %%mm2,  %%mm0
	 pfmul   %%mm3,  %%mm1
	 movq    %%mm0, (%%ecx)
	 movd    %%mm1,8(%%ecx)"
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]*b[0];
	c[1]=a[1]*b[1];
	c[2]=a[2]*b[2];

#endif

}

void simd_vecs3_recip(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b*/
	/* load up mmx registers */
	/* uses b as intermediate storage */
	/* perform the reciprocation */
	/* store the results */
	/* was full of bugs */
	
	__asm__ __volatile__(
	
    "movq     (%%eax), %%mm0
     movq      %%mm0,  %%mm1
	 punpckhdq %%mm1, %%mm1

	 movd    8(%%eax), %%mm2

	 pfrcp    %%mm0, %%mm4
	 pfrcpit1 %%mm4, %%mm0
	 pfrcpit2 %%mm4, %%mm0

	 pfrcp    %%mm1, %%mm5
	 pfrcpit1 %%mm5, %%mm1
	 pfrcpit2 %%mm5, %%mm1
	 
	 pfrcp    %%mm2, %%mm6
	 pfrcpit1 %%mm6, %%mm2
	 pfrcpit2 %%mm6, %%mm2
	 
	 movd %%mm0,  (%%ebx)
	 movd %%mm1, 4(%%ebx)
	 movd %%mm2, 8(%%ebx)"
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	b[0]=1/a[0];
	b[1]=1/a[1];
	b[2]=1/a[2];

#endif

}

/**********************************************************************/
/* Divide three floating-point numbers using 3DNow!                   */
/* *a , *b : pointers to arrays of 4 floats containing the input data */
/* *c      : pointer to array of 4 floats in which to store result    */
/**********************************************************************/

void simd_vecs3_div(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b*/
	/* load up mmx registers */
	/* uses b as intermediate storage */
	/* perform the reciprocation */
	/* store the results */
	/* was full of bugs! see previous function */
	
	__asm__ __volatile__(

    "movq     (%%ebx), %%mm0
     movq      %%mm0,  %%mm1
	 punpckhdq %%mm1,  %%mm1

	 movd 8(%%ebx), %%mm2

	 pfrcp    %%mm0, %%mm4
	 pfrcpit1 %%mm4, %%mm0
	 pfrcpit2 %%mm4, %%mm0

	 pfrcp    %%mm1, %%mm5
	 pfrcpit1 %%mm5, %%mm1
	 pfrcpit2 %%mm5, %%mm1
	 
	 pfrcp    %%mm2, %%mm6
	 pfrcpit1 %%mm6, %%mm2
	 pfrcpit2 %%mm6, %%mm2
	 
	 punpckldq %%mm1,  %%mm0
	 pfmul    (%%eax), %%mm0
	 movd    8(%%eax), %%mm3
	 pfmul     %%mm3,  %%mm2
	 movq      %%mm0, (%%ecx)
	 movd      %%mm2,8(%%ecx)"
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]/b[0];
	c[1]=a[1]/b[1];
	c[2]=a[2]/b[2];

#endif

}

/************************************************************************/
/* Assign floating point numbers a,b ,c and d to vector pointed to by v */
/* using 3DNow! (of any benefit?)                                       */
/************************************************************************/

void simd_vecs4_assign(float a, float b, float c, float d, float *v)
{


#ifdef USE_3DNOW

	/* set up the pointers : %eax = v, %edx = &a */
	/* move into MMX registers a,b and c,d*/
	/* assign to v */
	
	__asm__ __volatile__(

	"movd (%%eax), %%mm0
	 movd (%%ebx), %%mm2
	 movd (%%ecx), %%mm1
	 movd (%%edx), %%mm3

	 pfacc %%mm2, %%mm0
	 pfacc %%mm3, %%mm1
	 
	 movq %%mm0,  (%%edi)
	 movq %%mm1, 8(%%edi)"
	 :
	 : "a" (a), "b" (b), "c" (c), "d" (d), "D" (v)
	 : "eax", "memory");

#else

	/* in C : */
	v[0]=a;
	v[1]=b;
	v[2]=c;
	v[3]=d;

#endif

} 

/*******************************************/
/* Copy vector a into vetor b using 3DNow! */
/*******************************************/

void simd_vecs4_copy(float *a, float *b)
{

#ifdef USE_3DNOW

	/* set up the pointers : %eax = a, %edx = b */
	/* move elements of a into MMX registers */
	/* assign to b */
	__asm__ __volatile__(
	"movq  (%%eax),  %%mm0
	 movq 8(%%eax),  %%mm1
	 movq   %%mm0,  (%%ebx)
	 movq   %%mm1, 8(%%ebx)"
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* in C : */
	b[0]=a[0];
	b[1]=a[1];
	b[2]=a[2];
	b[3]=a[3];

#endif

}

/*********************************************************/
/* Sum two 4-element floating point vectors using 3DNow! */
/* *a , *b : pointers to vectors to be added             */
/* *c      : pointer to vector in which to store result  */
/*********************************************************/

void simd_vecs4_add(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = c, %edx = a, %ecx = b */
	/* load up mmx registers */
	/* perform the addition */
	/* store the results */
	
	__asm__ __volatile__(

    "movq   (%%eax),  %%mm0
	 movq  8(%%eax),  %%mm1
	 pfadd  (%%ebx),  %%mm0
	 pfadd 8(%%ebx),  %%mm1
	 movq    %%mm0,  (%%ecx)
	 movq    %%mm1, 8(%%ecx)"
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]+b[0];
	c[1]=a[1]+b[1];
	c[2]=a[2]+b[2];
	c[3]=a[3]+b[3];

#endif

}


/**************************************************************/
/* Subtract two 4-element floating point vectors using 3DNow! */
/* *a , *b : pointers to vectors to be subtracted             */
/* *c      : pointer to vector in which to store result       */
/**************************************************************/

void simd_vecs4_sub(float *a, float *b, float *c)
{

#ifdef USE_3DNOW

	/* set up pointers : %eax = c, %edx = a, %ecx = b */
	/* load up mmx registers */
	/* perform the addition */
	/* store the results */
	
	__asm__ __volatile__(
	"movq    (%%eax),  %%mm0
	 movq   8(%%eax),  %%mm1
	 pfsub   (%%ebx),  %%mm0
	 pfsub  8(%%ebx),  %%mm1
	 movq     %%mm0,  (%%ecx)
	 movq     %%mm1, 8(%%ecx)"
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]-b[0];
	c[1]=a[1]-b[1];
	c[2]=a[2]-b[2];
	c[3]=a[3]-b[3];

#endif

}


/**********************************************/
/* 14-10-2000 scalar product c=a.b, 4-vectors */
/**********************************************/
void simd_vecs4_scalar_product(float *a, float *b, float *c)
{

	#ifdef USE_3DNOW

	__asm__ __volatile(
	"movq   (%%eax), %%mm0
	 movq  8(%%eax), %%mm1
	 pfmul  (%%ebx), %%mm0
	 pfmul 8(%%ebx), %%mm1
	 pfadd   %%mm1,  %%mm0
	 pfacc   %%mm0,  %%mm0
	 movq    %%mm0, (%%ecx)
	"
	:
	: "a" (a), "b" (b), "c" (c)
	: "memory");

	#else

	c[0]=(a[0]*b[0])+(a[1]*b[1])+(a[2]*b[2])+(a[3]*b[3]);
	
	#endif

}

void simd_vecs4_mag(float *a, float *b)
{
#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = result */
	/* load up mmx registers */
	/* perform the operation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq      (%%eax), %%mm0
     movq     8(%%eax), %%mm1
     pfmul      %%mm0,  %%mm0
     pfmul      %%mm1,  %%mm1
     pfadd      %%mm1,  %%mm0
     pfacc      %%mm0,  %%mm0  
     pfsqrt     %%mm0,  %%mm1
     pfsqrtit1  %%mm1,  %%mm0
     pfsqrtit2  %%mm1,  %%mm0
     movd       %%mm0, (%%ebx)
     "
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	*b=sqrt((double)(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]+a[3]*a[3]));

#endif

}

/**********************************************************************/
/* Multiply eight floating-point numbers using 3DNow!                 */
/* *a , *b : pointers to arrays of 4 floats containing the input data */
/* *c      : pointer to array of 4 floats in which to store result    */
/**********************************************************************/

void simd_vecs4_mul(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = c, %edx = a, %ecx = b */
	/* load up mmx registers */
	/* perform the multiplication */
	/* store the results */
	
	__asm__ __volatile__(

    "movq   (%%eax),  %%mm0
	 movq  8(%%eax),  %%mm1
	 pfmul  (%%ebx),  %%mm0
	 pfmul 8(%%ebx),  %%mm1
	 movq    %%mm0,  (%%ecx)
	 movq    %%mm1, 8(%%ecx)"
	 :
	 : "a" (a), "b" (b), "c" (c)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]*b[0];
	c[1]=a[1]*b[1];
	c[2]=a[2]*b[2];
	c[3]=a[3]*b[3];

#endif

}

void simd_vecs4_recip(float *a, float *b)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b*/
	/* load up mmx registers */
	/* uses b as intermediate storage */
	/* perform the reciprocation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%eax), %%mm0
	 pfrcp %%mm0, %%mm1
	 movq %%mm0, %%mm4
	 pfrcpit1 %%mm1, %%mm0
	 movq 8(%%eax), %%mm2
	 pfrcpit2 %%mm1, %%mm0
	 movq %%mm0, (%%ebx)
	 pfrcp %%mm2, %%mm3
	 movq %%mm2, %%mm5
	 movl 4(%%eax), %%ecx
	 pfrcpit1 %%mm3, %%mm2
	 movl %%ecx, (%%eax)
	 pfrcpit2 %%mm3, %%mm2
	 movq (%%eax), %%mm0
	 pfrcp %%mm0, %%mm1
	 movq %%mm2, 8(%%ebx)
	 pfrcpit1 %%mm1, %%mm0
	 movl 12(%%eax), %%edx
	 pfrcpit2 %%mm1, %%mm0
	 movl %%edx, 8(%%eax)
	 movq 8(%%eax), %%mm2
	 pfrcp %%mm2, %%mm3
	 movl (%%ebx), %%ecx
	 pfrcpit1 %%mm3, %%mm2
	 movq %%mm0,(%%ebx)
	 pfrcpit2 %%mm3, %%mm2
	 movl %%ecx, (%%ebx)
	 movl 8(%%ebx), %%edx
	 movq %%mm2, 8(%%ebx)
	 movl %%edx, 8(%%ebx)
	 movq %%mm4, (%%eax)
	 movq %%mm5, 8(%%eax)"
	 :
	 : "a" (a), "b" (b)
	 : "memory");
#else

	/* this is what it does in C : */
	b[0]=1/a[0];
	b[1]=1/a[1];
	b[2]=1/a[2];
	b[3]=1/a[3];

#endif

}

/**********************************************************************/
/* Divide four floating-point numbers using 3DNow!                    */
/* *a , *b : pointers to arrays of 4 floats containing the input data */
/* *c      : pointer to array of 4 floats in which to store result    */
/**********************************************************************/

void simd_vecs4_div(float *a, float *b, float *c)
{

#ifdef USE_3DNOW
		
	/* set up pointers : %eax = a, %ebx = b*/
	/* load up mmx registers */
	/* uses b as intermediate storage */
	/* perform the reciprocation */
	/* store the results */
	
	__asm__ __volatile__(

    "movq  (%%eax), %%mm0
	 pfrcp %%mm0, %%mm1
	 movq %%mm0, %%mm4
	 movq 8(%%eax), %%mm2
	 pfrcpit1 %%mm1, %%mm0
	 pfrcpit2 %%mm1, %%mm0
	 movq %%mm0, (%%ebx)
	 pfrcp %%mm2, %%mm3
	 movq %%mm2, %%mm5
	 movl 4(%%eax), %%ecx
	 pfrcpit1 %%mm3, %%mm2
	 movl %%ecx, (%%eax)
	 pfrcpit2 %%mm3, %%mm2
	 movq (%%eax), %%mm0
	 pfrcp %%mm0, %%mm1
	 movq %%mm2, 8(%%ebx)
	 pfrcpit1 %%mm1, %%mm0
	 movl 12(%%eax), %%edx
	 pfrcpit2 %%mm1, %%mm0
	 movl %%edx, 8(%%eax)
	 movq 8(%%eax), %%mm2
	 pfrcp %%mm2, %%mm3
	 movl (%%ebx), %%ecx
	 pfrcpit1 %%mm3, %%mm2
	 movq %%mm0,(%%ebx)
	 pfrcpit2 %%mm3, %%mm2
	 movl %%ecx, (%%ebx)
	 movl 8(%%ebx), %%edx
	 movq %%mm2, 8(%%ebx)
	 movl %%edx, 8(%%ebx)
	 movq (%%ebx), %%mm0
	 pfmul (%%esi), %%mm0
	 movq 8(%%ebx), %%mm1
	 pfmul 8(%%esi), %%mm1
	 movq %%mm0, (%%ebx)
	 movq %%mm1, 8(%%ebx)
	 movq %%mm4, (%%eax)
	 movq %%mm5, 8(%%eax)"
	 :
	 : "a" (b), "b" (c), "S" (a)
	 : "memory");
#else

	/* this is what it does in C : */
	c[0]=a[0]/b[0];
	c[1]=a[1]/b[1];
	c[2]=a[2]/b[2];
	c[3]=a[3]/b[3];

#endif

}
